#ifndef VKF_INDUCTION_HPP
#define VKF_INDUCTION_HPP

#define DEFAULT_HOST "localhost"
#define DEFAULT_PASS "toor"
#define DEFAULT_USER "root"

#include <algorithm> // for std::sort
#include <functional> // for std::ref
#include <mutex> // for std::mutex
#include <thread> // for std::thread

#include "Hypothesis.hpp" // includes Obstacles.hpp, Fragment.hpp, Parents.hpp, and Candidate.hpp too
#include "Qualify.hpp"
#include <pybind11/pybind11.h>

#ifndef API
// `extern "C"` - specifies C linkage: forces the compiler to export function/variable by a pretty (unmangled) C name
#define API extern "C" BOOST_SYMBOL_EXPORT
#endif //API

typedef std::unique_ptr<std::thread> ThreadPtr;
typedef std::vector<ThreadPtr> ThreadsList; // std::vector admits push_emplace()

namespace vkf {

typedef std::unique_ptr<Hypothesis> HypothesisPtr;
typedef std::vector<HypothesisPtr> HypothesesList;

class Induction : public HypothesesList {
    protected:
    private:
        /**< Core data-members */
        std::unique_ptr<Sample> m_sample_ptr; //!< variable "m_sample" saves original training sample
        bool m_is_negative; //!< variable "m_is_negative" refers to the causes sign
        /**< Main data-members */
        AOMatrix m_fContext; //!< variable "m_fContext" saves formal context for the coupling Markov chain
        Obstacles m_obstacles; //!< variable "m_obstacle" saves list of opposite sign objects to forbid
        /**< Temporary data-members */
        unsigned int m_hypotheses_to_add; //!< variable "m_hypotheses_to_add" sets a number of added hypotheses
        unsigned int m_HypothesesNumber; //!< variable "m_HypothesesNumber" counts just added hypotheses
        /**< Thread-specific data-members */
        volatile bool m_StopRequested; //!< Member variable "m_StopRequest must be declared volatile"
        std::mutex m_ModificationMutex; //!< variable "m_ModificationMutex" blocks access to
        unsigned int m_ThreadsNumber; //!< variable "m_ThreadsNumber" saves the number of threads
        ThreadsList m_GenerationThreads; //!< variable "m_GenerationThreads must be the last data-member"
        /** \brief This is the static class function that serves as a C style pointer for the pthread_create call
         * All we do here is call the GenerateNext() function
         * \return none */
        static void* StartInduction(void* object_pointer) {
            return      reinterpret_cast<Induction*>(object_pointer)->GenerateNext();
        }
        /** \brief Generate and save next hypothesis
         * \return nullptr */
        void* GenerateNext();
    public:
        Induction() {
            m_sample_ptr = std::unique_ptr<Sample>(new Sample());
            m_is_negative = false;
        };

        // struct Hypotheses function members
        ~Induction() {
            Sample *sample_ptr = m_sample_ptr.release(); // release vkf::SamplePtr data member
            delete sample_ptr;
            // empty HypothesesList container
            while(!this->empty()) {
                this->pop_back();
            }
        };

        std::size_t LoadCHypotheses(    const Qualifier& thresholds,
                                        const Beget& generator,
                                        const std::string &sample_table_name,
                                        const std::string &hypotheses_table_name,
                                        const std::string &database,
                                        const std::string &db_host = DEFAULT_HOST,
                                        const std::string &db_user = DEFAULT_USER,
                                        const std::string &db_pswd = DEFAULT_PASS
                                    ) {
            this->clear();
            /**< Insert vkf::Sample  into vkf::Induction structures */ {
//                Sample sam(thresholds, generator, sample_table_name, database, db_host, db_user, db_pswd);
//                this->InsertSample(sam);
                this->InsertSample(Sample(thresholds, generator, sample_table_name, database, db_host, db_user, db_pswd));
            }
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            // obtain content of table with hypotheses
            mariadb::result_set_ref result = con->query("SHOW TABLES FROM " + database + " LIKE '" + hypotheses_table_name + "'");
            if (result->row_count() > 0) /* the table exists */ {
                /**< Load previously saved hypotheses from the table vkfhyps */
                result = con->query("SELECT * FROM " + hypotheses_table_name);
                // reading from a result set
                while (result->next()) {
                    bool suspicious = result->get_boolean("suspicious");
                    boost::dynamic_bitset<> encoded_bitset;
                    for (auto verges : thresholds) {
                        std::string attribute_name = verges.GetAttrName(); /**< This variable contains the name of an attribute in the current column */
                        std::string low_bound_name("low_");
                        low_bound_name.append(attribute_name);
                        double low_bound = result->get_double(low_bound_name); /**< This variable saves low bound on the value of the attribute */
                        std::string high_bound_name("high_");
                        high_bound_name.append(attribute_name);
                        double high_bound = result->get_double(high_bound_name); /**< This variable saves high bound on the value of the attribute */
                        boost::dynamic_bitset<> bdbs = verges.Encode(low_bound, high_bound);
                        for (std::size_t ndx = 0; ndx < bdbs.size(); ndx++ ) {
                            encoded_bitset.push_back(bdbs[ndx]);
                        }
                    }
                    this->emplace_back(std::make_unique<Hypothesis>(suspicious, std::move(encoded_bitset)));
                }
            }
            con->~connection();
            acc->~account();
            return this->size();
        };

        std::size_t LoadDHypotheses(    const Encoder& encoder,
                                        const std::string &sample_table_name,
                                        const std::string &hypotheses_table_name,
                                        const std::string &database,
                                        const std::string &db_host = DEFAULT_HOST,
                                        const std::string &db_user = DEFAULT_USER,
                                        const std::string &db_pswd = DEFAULT_PASS
                                    ) {
            this->clear();
            /**< Insert vkf::Sample  into vkf::Induction structures */ {
//                Sample sam(encoder, sample_table_name, database, db_host, db_user, db_pswd);
//                this->InsertSample(sam);
                this->InsertSample(Sample(encoder, sample_table_name, database, db_host, db_user, db_pswd));
            }
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            // obtain content of table with hypotheses
            mariadb::result_set_ref result = con->query("SHOW TABLES FROM " + database + " LIKE '" + hypotheses_table_name + "'");
            if (result->row_count() > 0) /* the table exists */ {
                /**< Load previously saved hypotheses from the table vkfhyps */
                result = con->query("SELECT * FROM " + hypotheses_table_name);
                // reading from a result set
                while (result->next()) {
                    bool suspicious = result->get_boolean("suspicious");
                    boost::dynamic_bitset<> encoded_bitset;
                    for (auto maps : encoder) {
                        std::string attribute_name = maps.getAttributeName(); /**< This variable contains the name of an attribute in the current column */
                        std::size_t pos = result->column_index(attribute_name);
                        std::string attribute_value = result->get_string(pos); /**< This variable saves the value of the attribute */
                        boost::dynamic_bitset<> bdbs = maps.findEncoding(attribute_value);
                        for (std::size_t ndx = 0; ndx < bdbs.size(); ndx++ ) {
                            encoded_bitset.push_back(bdbs[ndx]);
                        }
                    }
                    this->emplace_back(std::make_unique<Hypothesis>(suspicious, std::move(encoded_bitset)));
                }
            }
            con->~connection();
            acc->~account();
            return this->size();
        };

        std::size_t LoadFHypotheses(    const Encoder& encoder,
                                        const Qualifier& thresholds,
                                        const Beget& generator,
                                        const std::string &sample_table_name,
                                        const std::string &hypotheses_table_name,
                                        const std::string &database,
                                        const std::string &db_host = DEFAULT_HOST,
                                        const std::string &db_user = DEFAULT_USER,
                                        const std::string &db_pswd = DEFAULT_PASS
                                    ) {
            this->clear();
            /**< Insert vkf::Sample  into vkf::Induction structures */ {
//                Sample sam(encoder, sample_table_name, database, db_host, db_user, db_pswd);
//                this->InsertSample(sam);
                this->InsertSample(Sample(encoder, thresholds, generator, sample_table_name, database, db_host, db_user, db_pswd));
            }
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            // obtain content of table with hypotheses
            mariadb::result_set_ref result = con->query("SHOW TABLES FROM " + database + " LIKE '" + hypotheses_table_name + "'");
            if (result->row_count() > 0) /* the table exists */ {
                /**< Load previously saved hypotheses from the table vkfhyps */
                result = con->query("SELECT * FROM " + hypotheses_table_name);
                // reading from a result set
                while (result->next()) {
                    bool suspicious = result->get_boolean("suspicious");
                    boost::dynamic_bitset<> encoded_bitset;
                    for (auto maps : encoder) {
                        std::string attribute_name = maps.getAttributeName(); /**< This variable contains the name of an attribute in the current column */
                        std::size_t pos = result->column_index(attribute_name);
                        std::string attribute_value = result->get_string(pos); /**< This variable saves the value of the attribute */
                        boost::dynamic_bitset<> bdbs = maps.findEncoding(attribute_value);
                        for (std::size_t ndx = 0; ndx < bdbs.size(); ndx++ ) {
                            encoded_bitset.push_back(bdbs[ndx]);
                        }
                    }
                    for (auto verges : thresholds) {
                        std::string attribute_name = verges.GetAttrName(); /**< This variable contains the name of an attribute in the current column */
                        std::string low_bound_name("low_");
                        low_bound_name.append(attribute_name);
                        double low_bound = result->get_double(low_bound_name); /**< This variable saves low bound on the value of the attribute */
                        std::string high_bound_name("high_");
                        high_bound_name.append(attribute_name);
                        double high_bound = result->get_double(high_bound_name); /**< This variable saves high bound on the value of the attribute */
                        boost::dynamic_bitset<> bdbs = verges.Encode(low_bound, high_bound);
                        for (std::size_t ndx = 0; ndx < bdbs.size(); ndx++ ) {
                            encoded_bitset.push_back(bdbs[ndx]);
                        }
                    }
                    this->emplace_back(std::make_unique<Hypothesis>(suspicious, std::move(encoded_bitset)));
                }
            }
            con->~connection();
            acc->~account();
            return this->size();
        };

        bool InsertSample(Sample&& rhs) {
//        bool InsertSample(Sample& rhs) {
            bool is_correct = true;
            m_is_negative = false;
            m_sample_ptr = std::unique_ptr<Sample>(new Sample(std::move(rhs)));
            m_fContext = AOMatrix(*m_sample_ptr, m_is_negative); //vkf::AOMatrix::AOMatix(vkf::Sample&&, bool)
            m_obstacles = Obstacles(*m_sample_ptr, m_is_negative); //vkf::Obstacles::Obstacles(const vkf::Sample&, bool)
            // compute parents lists for prevously computed hypotheses
            if (this->size() > 0) {
                for (std::vector<HypothesisPtr>::const_iterator it = this->cbegin(); it != this->cend(); it++) {
                    Parents parents_list;
                    parents_list.resize(m_fContext.GetColumnLength(), false);
                    Fragment fragment;
                    fragment.swap((*it)->GetFragment());
                    m_fContext.Derive(fragment, /* result */ parents_list);
                    Fragment buffer;
                    buffer.resize(m_fContext.GetRowLength(), false);
                    m_fContext.Derive(parents_list, /* result */ buffer);
                    if (buffer != fragment) {
                        is_correct = false;
                        break;
                    } else  {
                        (*it)->SetParents(std::move(parents_list));
                        (*it)->SetFragment(std::move(fragment));
                    }
                }
                if (!is_correct) {
                    // empty HypothesesList container
                    while(!this->empty()) {
                        this->pop_back();
                    }
//                    std::cerr << "ERROR: wrong correspondence between sample and hypotheses!" << std::endl;
                    throw std::runtime_error("Wrong correspondence between sample and hypotheses");
                }
            }
            return is_correct;
        };

        std::size_t SaveCHypotheses(const Qualifier& encoder,
                                    const std::string &hypotheses_table_name,
                                    const std::string &database,
                                    const std::string &db_host = DEFAULT_HOST,
                                    const std::string &db_user = DEFAULT_USER,
                                    const std::string &db_pswd = DEFAULT_PASS
                                    ) {
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            con->execute("DROP TABLE IF EXISTS " + hypotheses_table_name);
            std::string sql_query("CREATE TABLE IF NOT EXISTS " + hypotheses_table_name);
            sql_query.append(" (row_id int(8) NOT NULL AUTO_INCREMENT, ");
            sql_query.append("suspicious tinyint(1) NOT NULL DEFAULT 1, ");
            sql_query.append("parents_number int(8) NOT NULL DEFAULT 1, ");
            std::string insert_query("INSERT INTO " + hypotheses_table_name + " VALUES (NULL, ?, ?");
            for (auto qualify : encoder) {
                sql_query.append("low_");
                sql_query.append(qualify.GetAttrName());
                sql_query.append(" float(53) NOT NULL DEFAULT 0.0, ");
                insert_query.append(", ?");
                sql_query.append("high_");
                sql_query.append(qualify.GetAttrName());
                sql_query.append(" float(53) NOT NULL DEFAULT 0.0, ");
                insert_query.append(", ?");
            }
            sql_query.append("PRIMARY KEY(row_id)) ");
            sql_query.append("ENGINE = InnoDB");
            insert_query.append(")");
            con->execute(sql_query);
            mariadb::statement_ref stmt = con->create_statement(insert_query);
            for (std::vector<HypothesisPtr>::const_iterator it = this->cbegin(); it != this->cend(); it++) {
                stmt->set_boolean(0, (*it)->IsSuspicious());
                stmt->set_unsigned32(1, (*it)->GetParentsNumber());
                Fragment fragment = (*it)->GetFragment();
                std::size_t left_point=0;
                for (std::uint32_t ndx = 0; ndx < encoder.size(); ndx++) {
                    // calculate a shift of the attribute at ndx
                    std::size_t add_length = 2*encoder[ndx].size();
                    // copy encoded attribute value into a buffer
                    boost::dynamic_bitset<> buffer;
                    for(std::size_t indx = 0; indx < add_length; indx++) {
                        buffer.push_back(fragment[left_point+indx]);
                    }
                    // compute a lower and an upper bounds on the attribbute value
                    std::pair<double,double> bounds = encoder[ndx].Decode(buffer);
                    stmt->set_double(2*ndx+2, bounds.first); // low_bound = bounds.first
                    stmt->set_double(2*ndx+3, bounds.second); // high_bound = bounds.second
                    left_point += add_length;
                }
                stmt->insert();
            }
//            stmt->~statement();
            con->~connection();
            acc->~account();
            return this->size();
        };

        std::size_t SaveDHypotheses(const Encoder& encoder,
                                    const std::string &hypotheses_table_name,
                                    const std::string &database,
                                    const std::string &db_host = DEFAULT_HOST,
                                    const std::string &db_user = DEFAULT_USER,
                                    const std::string &db_pswd = DEFAULT_PASS
                                    ) {
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            /**< Create the table vkfhyps to save computed hypotheses */
            // open database
            con->execute("USE " + database);
            con->execute("DROP TABLE IF EXISTS " + hypotheses_table_name);
            std::string sql_query("CREATE TABLE IF NOT EXISTS " + hypotheses_table_name);
            sql_query.append(" (row_id int(8) NOT NULL AUTO_INCREMENT, ");
            sql_query.append("suspicious tinyint(1) NOT NULL DEFAULT 1, ");
            sql_query.append("parents_number int(8) NOT NULL DEFAULT 1, ");
            std::string insert_query("INSERT INTO " + hypotheses_table_name + " VALUES (NULL, ?, ?");
            for (auto maps : encoder) {
                sql_query.append(maps.getAttributeName());
                sql_query.append(" varchar(255) NOT NULL DEFAULT 'null', ");
                insert_query.append(", ?");
            }
            sql_query.append("PRIMARY KEY(row_id)) ");
            sql_query.append("ENGINE = InnoDB");
            insert_query.append(")");
            con->execute(sql_query);
            mariadb::statement_ref stmt = con->create_statement(insert_query);
            for (std::vector<HypothesisPtr>::const_iterator it = this->cbegin(); it != this->cend(); it++) {
                stmt->set_boolean(0, (*it)->IsSuspicious());
                stmt->set_unsigned32(1, (*it)->GetParentsNumber());
                Fragment fragment = (*it)->GetFragment();
                std::size_t left_point=0;
                for (std::uint32_t ndx = 0; ndx < encoder.size(); ndx++) {
                    std::string attribute_name = encoder[ndx].getAttributeName();
                    // calculate a shift of the attribute at ndx
                    std::size_t add_length = encoder[ndx].findEncoding("null").size();
                    // copy encoded attribute value into a buffer
                    boost::dynamic_bitset<> buffer;
                    for(std::size_t indx = 0; indx < add_length; indx++) {
                        buffer.push_back(fragment[left_point+indx]);
                    }
                    // seek corresponding name of the value
                    stmt->set_string(ndx+2, encoder[ndx].findValueName(buffer));
                    left_point += add_length;
                }
                stmt->insert();
            }
//            stmt->~statement();
            con->~connection();
            acc->~account();
            return this->size();
        };

        std::size_t SaveFHypotheses(const Encoder& encoder,
                                    const Qualifier& thresholds,
                                    const std::string &hypotheses_table_name,
                                    const std::string &database,
                                    const std::string &db_host = DEFAULT_HOST,
                                    const std::string &db_user = DEFAULT_USER,
                                    const std::string &db_pswd = DEFAULT_PASS
                                    ) {
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            /**< Create the table vkfhyps to save computed hypotheses */
            // open database
            con->execute("USE " + database);
            con->execute("DROP TABLE IF EXISTS " + hypotheses_table_name);
            std::string sql_query("CREATE TABLE IF NOT EXISTS " + hypotheses_table_name);
            sql_query.append(" (row_id int(8) NOT NULL AUTO_INCREMENT, ");
            sql_query.append("suspicious tinyint(1) NOT NULL DEFAULT 1, ");
            sql_query.append("parents_number int(8) NOT NULL DEFAULT 1, ");
            std::string insert_query("INSERT INTO " + hypotheses_table_name + " VALUES (NULL, ?, ?");
            for (auto maps : encoder) {
                sql_query.append(maps.getAttributeName());
                sql_query.append(" varchar(255) NOT NULL DEFAULT 'null', ");
                insert_query.append(", ?");
            }
            for (auto qualify : thresholds) {
                sql_query.append("low_");
                sql_query.append(qualify.GetAttrName());
                sql_query.append(" float(53) NOT NULL DEFAULT 0.0, ");
                insert_query.append(", ?");
                sql_query.append("high_");
                sql_query.append(qualify.GetAttrName());
                sql_query.append(" float(53) NOT NULL DEFAULT 0.0, ");
                insert_query.append(", ?");
            }
            sql_query.append("PRIMARY KEY(row_id)) ");
            sql_query.append("ENGINE = InnoDB");
            insert_query.append(")");
            con->execute(sql_query);
            mariadb::statement_ref stmt = con->create_statement(insert_query);
            for (std::vector<HypothesisPtr>::const_iterator it = this->cbegin(); it != this->cend(); it++) {
                stmt->set_boolean(0, (*it)->IsSuspicious());
                stmt->set_unsigned32(1, (*it)->GetParentsNumber());
                Fragment fragment = (*it)->GetFragment();
                std::size_t left_point=0;
                for (std::uint32_t ndx = 0; ndx < encoder.size(); ndx++) {
                    std::string attribute_name = encoder[ndx].getAttributeName();
                    // calculate a shift of the attribute at ndx
                    std::size_t add_length = encoder[ndx].findEncoding("null").size();
                    // copy encoded attribute value into a buffer
                    boost::dynamic_bitset<> buffer;
                    for(std::size_t indx = 0; indx < add_length; indx++) {
                        buffer.push_back(fragment[left_point+indx]);
                    }
                    // seek corresponding name of the value
                    stmt->set_string(ndx+2, encoder[ndx].findValueName(buffer));
                    left_point += add_length;
                }
                for (std::uint32_t ndx = 0; ndx < thresholds.size(); ndx++) {
                    // calculate a shift of the attribute at ndx
                    std::size_t add_length = 2*thresholds[ndx].size();
                    // copy encoded attribute value into a buffer
                    boost::dynamic_bitset<> buffer;
                    for(std::size_t indx = 0; indx < add_length; indx++) {
                        buffer.push_back(fragment[left_point+indx]);
                    }
                    // compute a lower and an upper bounds on the attribbute value
                    std::pair<double,double> bounds = thresholds[ndx].Decode(buffer);
                    stmt->set_double(2*ndx+encoder.size()+2, bounds.first); // low_bound = bounds.first
                    stmt->set_double(2*ndx+encoder.size()+3, bounds.second); // high_bound = bounds.second
                    left_point += add_length;
                }
                stmt->insert();
            }
//            stmt->~statement();
            con->~connection();
            acc->~account();
            return this->size();
        };

        /** Inductive generalization */
        std::size_t AddHypotheses(unsigned int hypotheses_number, unsigned int threads_number) {
            m_HypothesesNumber = 0;
            m_hypotheses_to_add = hypotheses_number;
            m_StopRequested = false;
            m_ThreadsNumber = threads_number;
            for(std::size_t i = 0; i < m_ThreadsNumber; ++i) {
                m_GenerationThreads.emplace_back(ThreadPtr(new std::thread(std::bind(Induction::StartInduction, this))));
            }
            while(!m_GenerationThreads.empty()) {
                m_GenerationThreads.back()->join();
                m_GenerationThreads.pop_back();
            }
            return this->size();
        }

        /** Garbage hypotheses detection declaration */
        unsigned int DetectSuspicious();
        /** Garbage hypotheses restore status */
        void RestoreStatus();

        const unsigned int GetSize() const {
            return size();
        };

         bool IsPositive(const boost::dynamic_bitset<>& test) const {
            if (size() == 0) {
                return false;
            }
            if ((this->front())->GetFragment().size() != test.size()) {
                return false;
            }
            bool result = false;
            for (HypothesesList::const_iterator cit = this->cbegin(); cit != this->cend(); cit++) {
                if ((*cit)->GetFragment().is_subset_of(test) && !(*cit)->IsSuspicious()) {
                    result = true;
                    break;
                }
            }
            return result;
        }

        const unsigned int GetParentsNumber(unsigned int ndx) const {return (this->at(ndx))->GetParentsNumber();};

         Fragment& GetFragment(unsigned int ndx) const {return (this->at(ndx))->GetFragment();};

        /** ShowCHypothesis function implementation
         * \param thresholds is used to decode a hypothesis at position ndx
         * \param ndx determines a position of the hypothesis to be displayed
         * \return pybind11::list of nontrivial pairs <attribute/value> */
        pybind11::list ShowCHypothesis(const Qualifier &thresholds, const std::size_t ndx) const {
            pybind11::list result;
            std::size_t add_length;
            std::string feature_name;
            std::pair<double,double> value_interval;
            std::size_t left_point = 0;
            for (auto part : thresholds) {
                add_length = 2*part.size();
                // copy encoded attribute value into a buffer
                boost::dynamic_bitset<> buffer;
                for(std::size_t indx = 0; indx < add_length; ++indx) {
                    buffer.push_back(GetFragment(ndx).test(left_point + indx));
                }
                // seek for names of the feature and corresponding value
                feature_name = part.GetAttrName();
                value_interval = part.Decode(buffer);
                result.append(pybind11::make_tuple(feature_name, value_interval));
                left_point += add_length;
            }
            return result;
        }

        /** ShowDHypothesis function implementation
         * \param encoder is used to decode a hypothesis at position ndx
         * \param ndx determines a position of the hypothesis to be displayed
         * \return pybind11::list of nontrivial pairs <attribute/value> */
        pybind11::list ShowDHypothesis(const Encoder& encoder, const std::size_t ndx) const {
            pybind11::list result;
            std::size_t add_length;
            std::string feature_name;
            std::string value_name;
            std::size_t left_point = 0;
            for (auto part : encoder) {
                add_length = part.findEncoding("null").size();
                // copy encoded attribute value into a buffer
                boost::dynamic_bitset<> buffer;
                for(std::size_t indx = 0; indx < add_length; ++indx) {
                    buffer.push_back(GetFragment(ndx).test(left_point + indx));
                }
                // seek for names of the feature and corresponding value
                feature_name = part.getAttributeName();
                value_name = part.findValueName(buffer);
                if (strcmp(value_name.c_str(), "null") != 0) {
                    result.append(pybind11::make_tuple(feature_name, value_name));
                }
                left_point += add_length;
            }
            return result;
        }

        /** ShowFHypothesis function implementation
         * \param encoder is used to decode a hypothesis at position ndx
         * \param thresholds is used to decode a hypothesis at position ndx
         * \param ndx determines a position of the hypothesis to be displayed
         * \return pybind11::list of nontrivial pairs <attribute/value> */
        pybind11::list ShowFHypothesis(const Encoder& encoder, const Qualifier &thresholds, const std::size_t ndx) const {
            pybind11::list result;
            std::size_t add_length;
            std::string feature_name;
            std::string value_name;
            std::pair<double,double> value_interval;
            std::size_t left_point = 0;
            for (auto part : encoder) {
                add_length = part.findEncoding("null").size();
                // copy encoded attribute value into a buffer
                boost::dynamic_bitset<> buffer;
                for(std::size_t indx = 0; indx < add_length; ++indx) {
                    buffer.push_back(GetFragment(ndx).test(left_point + indx));
                }
                // seek for names of the feature and corresponding value
                feature_name = part.getAttributeName();
                value_name = part.findValueName(buffer);
                if (strcmp(value_name.c_str(), "null") != 0) {
                    result.append(pybind11::make_tuple(feature_name, value_name));
                }
                left_point += add_length;
            }
            for (auto part : thresholds) {
                add_length = 2*part.size();
                // copy encoded attribute value into a buffer
                boost::dynamic_bitset<> buffer;
                for(std::size_t indx = 0; indx < add_length; ++indx) {
                    buffer.push_back(GetFragment(ndx).test(left_point + indx));
                }
                // seek for names of the feature and corresponding value
                feature_name = part.GetAttrName();
                value_interval = part.Decode(buffer);
                result.append(pybind11::make_tuple(feature_name, value_interval));
                left_point += add_length;
            }
            return result;
        }

};
} // namespace vkf

#endif // VKF_INDUCTION_HPP
