#include "Inhabit.hpp"

bool vkf::pair_compare( std::pair<double, bool> const &pair1, std::pair<double, bool> const &pair2 ) {
    return (pair1.first < pair2.first);
};

double vkf::entropy(std::size_t pos, std::size_t neg) {
    if (pos > 0 && neg > 0) {
        return (-std::log2((double)neg/(pos+neg))*neg/(pos+neg)-std::log2((double)pos/(pos+neg))*pos/(pos+neg));
    } else {
        return 0;
    }
};

double vkf::gain(std::size_t left_pos, std::size_t left_neg, std::size_t right_pos, std::size_t right_neg) {
    return entropy(left_pos, left_neg)*(left_pos+left_neg)/(left_pos+left_neg+right_pos+right_neg) +
        entropy(right_pos, right_neg)*(right_pos+right_neg)/(left_pos+left_neg+right_pos+right_neg);
};

/** ComputeTheshold detects the best entropy verge for sorted values of continuous attribute
 * \param begin_it sets the start point to search the best verge
 * \param end_it determines a position to stot search (an element on this position is ignored)
 * \return double value of the best theshold */
double vkf::Inhabit::ComputeVerge(std::list<std::pair<double, bool>>::iterator begin_it, std::list<std::pair<double, bool>>::iterator end_it) {
    std::size_t left_pos = 0;
    std::size_t left_neg = 0;
    std::size_t right_pos = 0;
    std::size_t right_neg = 0;
    auto candidate_it = std::find_if(begin_it, end_it,
        [&](const std::pair<double,bool> &arg){ return (arg.first > begin_it->first); });
    double verge_result = candidate_it->first;
    for(auto it = begin_it; it != candidate_it; it++) {
        if (it->second) {
            left_pos++;
        } else {
            left_neg++;
        }
    }
    for(auto it = candidate_it; it != end_it; it++) {
        if (it->second) {
            right_pos++;
        } else {
            right_neg++;
        }
    }
    double best_result = vkf::gain(left_pos, left_neg, right_pos, right_neg);
    for (auto it = candidate_it; it != end_it; ) {
        std::list<std::pair<double, bool>>::iterator next_it = it;
        while (next_it->first == it->first) {
            if (next_it->second) {
                right_pos--;
                left_pos++;
            } else {
                right_neg--;
                left_neg++;
            }
            next_it++;
        }
        double result = vkf::gain(left_pos, left_neg, right_pos, right_neg);
        if (next_it != end_it && best_result > result) {
            best_result = result;
            verge_result = next_it->first;
        }
        it = next_it;
    }
    return verge_result;
};

/** ComputeEntropy computes the entropy of the values from begin_it to end_it
 * \param begin_it sets the start point
 * \param end_it determines a position to stop search
 * \return double value of the entropy from begin_it to end_it */
double vkf::Inhabit::ComputeEntropy(std::list<std::pair<double, bool>>::iterator begin_it, std::list<std::pair<double, bool>>::iterator end_it) {
    std::size_t pos_total = 0;
    std::size_t neg_total = 0;
    for(auto it = begin_it; it != end_it; it++) {
        if (it->second) {
            pos_total++;
        } else {
            neg_total++;
        }
    }
    return vkf::entropy(pos_total, neg_total);
};

/** ComputeLeftEntropy computes the entropy of the values from begin_it to the verge
 * \param begin_it sets the start point to search the best verge
 * \param end_it determines a position to stop the search (an element on this position is ignored)
 * \return double value of the entropy from begin_it to ComputeVerge() */
double vkf::Inhabit::ComputeLeftEntropy(std::list<std::pair<double, bool>>::iterator begin_it, std::list<std::pair<double, bool>>::iterator end_it) {
    double verge_result = vkf::Inhabit::ComputeVerge(begin_it, end_it);
    std::size_t left_pos = 0;
    std::size_t left_neg = 0;
    std::list<std::pair<double,bool>>::iterator stop_it = std::find_if(begin_it, end_it, [&](const std::pair<double,bool> &arg){ return arg.first == verge_result; } );
    for(auto it = begin_it; it != stop_it; it++) {
        if (it->second) {
            left_pos++;
        } else {
            left_neg++;
        }
    }
    return vkf::entropy(left_pos, left_neg);
};

/** ComputeRightEntropy computes the entropy of the values from the verge to the end_it
 * \param begin_it sets the start point to search the best verge
 * \param end_it determines a position to stop the search (an element on this position is ignored)
 * \return double value of the entropy from ComputeVerge() to the end_it */
double vkf::Inhabit::ComputeRightEntropy(std::list<std::pair<double, bool>>::iterator begin_it, std::list<std::pair<double, bool>>::iterator end_it) {
    double verge_result = vkf::Inhabit::ComputeVerge(begin_it, end_it);
    std::size_t right_pos = 0;
    std::size_t right_neg = 0;
    std::list<std::pair<double,bool>>::iterator start_it = std::find_if(begin_it, end_it, [&](const std::pair<double,bool> &arg){ return arg.first == verge_result; } );
    for(auto it = start_it; it != end_it; it++) {
        if (it->second) {
            right_pos++;
        } else {
            right_neg++;
        }
    }
    return vkf::entropy(right_pos, right_neg);
};
