#ifndef VKF_MCSTATE_HPP
#define VKF_MCSTATE_HPP

#include "Candidate.hpp" // Classes AOMatrix and Candidates declarations, it includes boostdll.h, general.h, Fragment.hpp, Parents.hpp, and Encoded.hpp too

namespace vkf {
    class MCState {
        protected:
        private:
            CandidatePtr pointer_to_upper_candidate; //!< Pointer to associated upper Candidate object
	        CandidatePtr pointer_to_lower_candidate; //!< Pointer to associated lower Candidate object
            bool moveUp; //!< Does the previous move direction be up (or down)?
        public:
            /** \brief Default constructor declaration
            * \param fContext contains training sample to make Markov chain over pairs of candidates */
            MCState(const AOMatrix& context) {
                moveUp = false;
                pointer_to_upper_candidate = std::move(std::unique_ptr<Candidate>(new Candidate(context, CandidateType::TOP_CANDIDATE)));
                pointer_to_lower_candidate = std::move(std::unique_ptr<Candidate>(new Candidate(context, CandidateType::BOTTOM_CANDIDATE)));
            };

            ~MCState() = default; /**< Default destructor implementation */

            /** \brief The Markov chain step by application of operation CloseByOne to both components
             * \param fContext contains training sample to make Markov chain over pairs of candidates
             * \return the modified state of the Markov chain */
            void MoveBothComponents(const AOMatrix& context) {
                std::size_t random_index = AOMatrix::index_distribution(AOMatrix::mt_random_engine);
                bool next_move_up = context.IsNextMoveUp(random_index);
                if (next_move_up != moveUp) {
                    this->Complete(context);
                }
                if (next_move_up) {
                    random_index -= context.GetColumnLength();
                    if (pointer_to_upper_candidate->GetAttributes()[random_index] == false) {
                        context.CbOUp(pointer_to_upper_candidate->GetObjects(), random_index);
                    }
                    if (pointer_to_lower_candidate->GetAttributes()[random_index] == false) {
                        context.CbOUp(pointer_to_lower_candidate->GetObjects(), random_index);
                    }
                }
                else /* next_move_up == false */ {
                    if (pointer_to_upper_candidate->GetObjects()[random_index] == false) {
                        context.CbODown(pointer_to_upper_candidate->GetAttributes(), random_index);
                    }
                    if (pointer_to_lower_candidate->GetObjects()[random_index] == false) {
                        context.CbODown(pointer_to_lower_candidate->GetAttributes(), random_index);
                    }
                }
                moveUp = next_move_up;
            };

            /** \brief Complete the Markov chain computation
             * \param context contains training sample to make Markov chain over pairs of candidates
             * \return nothing */
            void Complete(const AOMatrix& context) const {
                pointer_to_upper_candidate->Complete(context, moveUp);
                pointer_to_lower_candidate->Complete(context, moveUp);
                return;
            };

            /** \brief IsCoupled determines a termination of the Markov chain computation
             * \return true if Markov chain computation finished */
            bool IsCoupled() const {
                if (moveUp) {
                    return (pointer_to_upper_candidate->GetObjects() == pointer_to_lower_candidate->GetObjects());
                }
                else /* moveUp == false */ {
                    return (pointer_to_upper_candidate->GetAttributes() == pointer_to_lower_candidate->GetAttributes());
                }
            };

            /** Access to the upper Candidate
             * \return the current value of the upper Candidate */
            Candidate&& GetUpperCandidate() const {return std::move(*pointer_to_upper_candidate);}

    };
} //namespace vkf

#endif // VKF_MCSTATE_HPP
